/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlHidden.h
// = AUTHOR(S)
//     Marco Jacques, Aghiles Kheffache
// = VERSION
//     $Revision: 1.10 $
// = DATE RELEASED
//     $Date: 2003/01/15 20:31:43 $
// = RCSID
//     $Id: dlHidden.h,v 1.10 2003/01/15 20:31:43 aghiles Exp $
// ===========================================================================

#ifndef __dlHidden_H
#define __dlHidden_H

#include "dlTokenInfo.h"
#include "dlVar.h"
#include "dlBasicTypes.h"

class DlTokenRegWithDefault;

/*
  class DlHidden

  Abstract parent class used by a grid (or anything else that can be shaded) to
  store values for hidden parameters.  Subclasses will define the shader 
  specific variables.
*/
class DlHidden
{
public:

	DlHidden();
	virtual ~DlHidden();

	// enumerates ALL the possible shader predefined parameters
	enum hiddenNames
	{
		/* bad */
		e_null = 0x00000000,

		/* for surface & displacement only */
		e_Cs  =0x00000001,
		e_Os  =0x00000002,
		e_Ci  =0x00000004,
		e_Oi  =0x00000008,
		e_I   =0x00000010,

		/* for all shaders */
		e_E   =0x00000020,
		e_P   =0x00000040,
		e_u   =0x00000080,
		e_v   =0x00000100,
		e_du  =0x00000200,
		e_dv  =0x00000400,
		e_dPdu=0x00000800,
		e_dPdv=0x00001000,
		e_s   =0x00002000,
		e_t   =0x00004000,
		e_N   =0x00008000,
		e_Ng  =0x00010000,

		/* for light shaders only */
		e_Ps  =0x00020000,
		e_L   =0x00040000,
		e_Cl  =0x00080000,
		e_Ol  =0x00100000,

		/* for image shaders only */
		e_alpha  =0x00200000,
		e_ncomps =0x00400000,
		e_Z      =0x00800000,

		/* time */
		e_time   =0x01000000
	};

	/*
		WARNING: when adding/removing a predefined variable in the enumaration
		above, do not forget to increment this count.
	*/
	enum { k_numVariables = 26 };

	static hiddenNames Name( unsigned i );
	static hiddenNames Name( const char * );

	/*
		NeededMask

		returns the mask of needed variable
	*/
	unsigned NeededMask( void ) const { return m_neededMask; }

	/* static unsigned Mask( hiddenNames i_name ) { (unsigned) i_name; }
	*/
	
	virtual void Init( const DlTokenRegWithDefault *, unsigned i_neededMask );

	virtual DlVar *GetVar( hiddenNames i_name, DlVar ***o_dst=0x0 ) const;

	/*
		Infos

		returns useful informations about some variable number.
	*/
	/* bool Infos(
		unsigned i_num,
		DlVar **o_var, hiddenNames *o_name, unsigned *o_mask ) const;
		*/

	/*
		Get

		gets the value of some named parameter

		PARAMETERS
		i_name    : parameter ID
	*/
	float *Get( hiddenNames i_name ) const;

	/*
		NumChannels

		returns the number of channels used by some named parameter

		PARAMETERS
		i_name    : parameter name
	*/
	unsigned NumChannels( hiddenNames i_name ) const
	{
		DlVar *v = GetVar( i_name );
		return v ? v->NumChannels() : 0;
	}

	/*
		NumValues

		returns number of values contained in some name variable

		NOTE
		for uniform & constant variables the return value should be 1, for
		varying variables the return value should be at least 4.
	*/
	unsigned NumValues( hiddenNames i_name )
	{
		DlVar *v = GetVar( i_name );
		return v ? v->NumValues() : 0;
	}

	/*
		GetVarInfos

		returns some variables value, numChannels & numValues.
	*/
	bool GetVarInfos(
		hiddenNames i_name, float **o_value,
		unsigned *o_numChans, unsigned *o_numVals ) const;

	/*
		Set (1)

		sets the value of some named parameter

		PARAMETERS
		i_name    : parameter name
	*/
	bool Set(
		hiddenNames i_name,
		const DlTokenInfo::DlClass& i_valueClass,
		unsigned i_numValues, void* i_value ) const;

	/*
		NumParameters

		returns number of total number of standard SL parameters
	*/
	static unsigned NumParameters( void )
	{
		return k_numVariables;
	}

public:

	// total size of this structure along with its members
	mutable unsigned m_size;

	DlVector3* GetP(void) const { return m_P ? (DlVector3*)m_P->Value() : 0x0; }
	DlVector3* GetDPdu(void) const { return m_dPdu ? (DlVector3*)m_dPdu->Value() : 0x0; }
	DlVector3* GetDPdv(void) const { return m_dPdv ? (DlVector3*)m_dPdv->Value() : 0x0; }
	DlVector3* GetN(void) const { return m_N ? (DlVector3*)m_N->Value() : 0x0; }
	DlVector3* GetNg(void) const { return m_Ng ? (DlVector3*)m_Ng->Value() : 0x0; }

	float* GetU(void) const { return m_u ? (float*)m_u->Value() : 0x0; }
	float* GetV(void) const { return m_v ? (float*)m_v->Value() : 0x0; }
	float* GetDu(void) const { return m_du ? (float*)m_du->Value() : 0x0; }
	float* GetDv(void) const { return m_dv ? (float*)m_dv->Value() : 0x0; }
	float* GetS(void) const { return m_s ? (float*)m_s->Value() : 0x0; }
	float* GetT(void) const { return m_t ? (float*)m_t->Value() : 0x0; }
	float* GetE(void) const { return m_E ? (float*)m_E->Value() : 0x0; }
	float* GetNcomps(void) const { return m_ncomps ? (float*)m_ncomps->Value() : 0x0; }
	float* GetTime(void) const { return m_time ? (float*)m_time->Value() : 0x0; }	

	void SetP(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetDPdu(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetDPdv(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetN(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetNg(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );

	void SetU(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetV(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetDu(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetDv(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetS(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetT(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetE(const DlTokenInfo::DlClass&, unsigned i_numValues, void* );

	void SetNcomps( const DlTokenInfo::DlClass&, unsigned i_numValues, void* );
	void SetTime( const DlTokenInfo::DlClass&, unsigned i_numValues, void* );

protected:

	/*
		Set (2)

		Sets some variable to a given value

		PARAMETERS
		i_name    : variable's name
		i_var     : source variable

		NOTE
		- the destination variable 'i_name' should be 0x0
		- it is the duty of this class to deallocate i_var ltr, this is not
		super clean and that is why this method is private!
	*/
	void Set( hiddenNames i_name, DlVar *i_var );

private:
	DlVar* m_P;
	DlVar* m_dPdu;
	DlVar* m_dPdv;
	DlVar* m_N;
	DlVar* m_Ng;

	DlVar* m_u;
	DlVar* m_v;
	DlVar* m_du;
	DlVar* m_dv;
	DlVar* m_s;
	DlVar* m_t;	

	DlVar *m_E;

	DlVar *m_ncomps;
	DlVar *m_time;

	unsigned m_neededMask;
};

#endif
