/******************************************************************************/
/*                                                                            */
/*    Copyright (c) Centre de traitement en imagerie virtuelle inc.           */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/*    These coded instructions, statements, and computer programs  contain    */
/*    unpublished  proprietary and  confidential information of  Centre de    */
/*    traitement  en  imagerie virtuelle inc. (CTIV), and are protected by    */
/*    Federal copyright law. They may not be disclosed to third parties or    */
/*    or duplicated in any form, in whole or in  part, without  the  prior    */
/*    written consent of CTIV.                                                */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlMatrix44.h
// = AUTHOR(S)
//     Franck Diard, Patrick Fournier
// = VERSION
//     $Revision: 1.1 $
// = DATE RELEASED
//     $Date: 2000/06/21 18:46:36 $
// = RCSID
//     $Id: dlMatrix44.h,v 1.1 2000/06/21 18:46:36 ours Exp $
// = COPYRIGHT
//     Copyright  1999 Centre de traitement en imagerie virtuelle inc.
//     All Rights Reserved.
// ===========================================================================

#ifndef __dlMatrix44_H
#define __dlMatrix44_H

#include <string.h>
#include "dlVector3.h"

/*
  Matrix indices:

    0  1  2  3
    4  5  6  7
    8  9 10 11
   12 13 14 15
*/


#ifdef DELIGHT_STATS

#ifndef DL_COMPILING_SHADER
__declspec(dllexport) extern unsigned long g_numMatrixInversions;
__declspec(dllexport) extern unsigned long g_numMatrixMuls_Point;
__declspec(dllexport) extern unsigned long g_numMatrixMuls_Vector;
__declspec(dllexport) extern unsigned long g_numMatrixMuls_Normal;
#endif

#endif

void Matrix44PrintStats();

typedef float DlMatrix44[16];
typedef float DlBasis44[16];

static const DlMatrix44 s_identityMatrix =
{
	1.0, 0.0, 0.0, 0.0,
	0.0, 1.0, 0.0, 0.0,
	0.0, 0.0, 1.0, 0.0,
	0.0, 0.0, 0.0, 1.0
};

void Matrix44ByNormal(DlVector3& o_normal, const DlMatrix44& i_m, const DlVector3& i_normal);
void Matrix44ByNormalWithW(DlVector3& o_normal, const DlMatrix44& i_m, const DlVector3& i_normal);

void Matrix44Invert(const DlMatrix44 in, DlMatrix44 out);
double Matrix44Determinant(const DlMatrix44 m);

/* Make a projection matrix from the field of view specified.
   If 0 > i_fieldOfView > 180, the projection matrix is a perspective one,
   else it is an orthographic one.
*/

void Matrix44BuildTranslationMatrix(DlMatrix44 o_mat, const DlVector3& i_t);
void Matrix44BuildRotationMatrix(DlMatrix44 o_mat, float i_angle, const DlVector3& i_axis);
void Matrix44BuildScaleMatrix( DlMatrix44 o_mat, float sx, float sy, float sz );
void Matrix44BuildSkewMatrix( DlMatrix44 o_mat, float angle, const DlVector3& i_axis1, const DlVector3& i_axis2 );
void Matrix44BuildProjectionMatrix(DlMatrix44 out, float i_fieldOfView);
void Matrix44BuildPseudoPerspectiveMatrix(DlMatrix44 o_mat, float i_fieldOfView);
void Matrix44BuildPseudoPerspectiveInvertMatrix(DlMatrix44 o_mat, float i_fieldOfView);


#define FLOAT_EQUAL(R1, R2, EPSILON) (fabs((R1) - (R2)) <= EPSILON)
#define FLOAT_STRICTLY_EQUAL(R1, R2) ((R1) == (R2))

#define MATRIX44_SET(Z, \
	A, B, C, D, \
	E, F, G, H, \
	I, J, K, L, \
	M, N, P, Q) \
	{ \
		(Z)[ 0] = (A); (Z)[ 1] = (B); (Z)[ 2] = (C); (Z)[ 3] = (D);\
		(Z)[ 4] = (E); (Z)[ 5] = (F); (Z)[ 6] = (G); (Z)[ 7] = (H);\
		(Z)[ 8] = (I); (Z)[ 9] = (J); (Z)[10] = (K); (Z)[11] = (L);\
		(Z)[12] = (M); (Z)[13] = (N); (Z)[14] = (P); (Z)[15] = (Q);\
	}

#define MATRIX44_ELEMENT(M, L, C)       (M[(L)*4+(C)])

/* Copy M2 into M1 */
#define MATRIX44_COPY(M1, M2)\
	{ \
		memcpy((M1), (M2), sizeof(DlMatrix44)); \
	}

#define MATRIX44_IDENTITY(M)\
	{ \
		MATRIX44_COPY((M), s_identityMatrix); \
	}

#define MATRIX44_EQUAL(M1, M2, EPSILON)\
	( \
		FLOAT_EQUAL((M1)[ 0], (M2)[ 0], EPSILON) && \
		FLOAT_EQUAL((M1)[ 1], (M2)[ 1], EPSILON) && \
		FLOAT_EQUAL((M1)[ 2], (M2)[ 2], EPSILON) && \
		FLOAT_EQUAL((M1)[ 3], (M2)[ 3], EPSILON) && \
		FLOAT_EQUAL((M1)[ 4], (M2)[ 4], EPSILON) && \
		FLOAT_EQUAL((M1)[ 5], (M2)[ 5], EPSILON) && \
		FLOAT_EQUAL((M1)[ 6], (M2)[ 6], EPSILON) && \
		FLOAT_EQUAL((M1)[ 7], (M2)[ 7], EPSILON) && \
		FLOAT_EQUAL((M1)[ 8], (M2)[ 8], EPSILON) && \
		FLOAT_EQUAL((M1)[ 9], (M2)[ 9], EPSILON) && \
		FLOAT_EQUAL((M1)[10], (M2)[10], EPSILON) && \
		FLOAT_EQUAL((M1)[11], (M2)[11], EPSILON) && \
		FLOAT_EQUAL((M1)[12], (M2)[12], EPSILON) && \
		FLOAT_EQUAL((M1)[13], (M2)[13], EPSILON) && \
		FLOAT_EQUAL((M1)[14], (M2)[14], EPSILON) && \
		FLOAT_EQUAL((M1)[15], (M2)[15], EPSILON) && \
	)

#define MATRIX44_STRICTLY_EQUAL(M1, M2)\
	( \
		FLOAT_STRICTLY_EQUAL((M1)[ 0], (M2)[ 0]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 1], (M2)[ 1]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 2], (M2)[ 2]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 3], (M2)[ 3]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 4], (M2)[ 4]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 5], (M2)[ 5]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 6], (M2)[ 6]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 7], (M2)[ 7]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 8], (M2)[ 8]) && \
		FLOAT_STRICTLY_EQUAL((M1)[ 9], (M2)[ 9]) && \
		FLOAT_STRICTLY_EQUAL((M1)[10], (M2)[10]) && \
		FLOAT_STRICTLY_EQUAL((M1)[11], (M2)[11]) && \
		FLOAT_STRICTLY_EQUAL((M1)[12], (M2)[12]) && \
		FLOAT_STRICTLY_EQUAL((M1)[13], (M2)[13]) && \
		FLOAT_STRICTLY_EQUAL((M1)[14], (M2)[14]) && \
		FLOAT_STRICTLY_EQUAL((M1)[15], (M2)[15])    \
	)

#define MATRIX44_MULTIPLY_IN_PLACE(M1, M2)\
	{ \
		DlMatrix44 t_; \
			MATRIX44_COPY(t_, (M1)); \
			(M1)[ 0] = t_[ 0]*(M2)[ 0]+t_[ 1]*(M2)[ 4]+t_[ 2]*(M2)[ 8]+t_[ 3]*(M2)[ 12]; \
			(M1)[ 1] = t_[ 0]*(M2)[ 1]+t_[ 1]*(M2)[ 5]+t_[ 2]*(M2)[ 9]+t_[ 3]*(M2)[ 13]; \
			(M1)[ 2] = t_[ 0]*(M2)[ 2]+t_[ 1]*(M2)[ 6]+t_[ 2]*(M2)[10]+t_[ 3]*(M2)[ 14]; \
			(M1)[ 3] = t_[ 0]*(M2)[ 3]+t_[ 1]*(M2)[ 7]+t_[ 2]*(M2)[11]+t_[ 3]*(M2)[ 15]; \
			(M1)[ 4] = t_[ 4]*(M2)[ 0]+t_[ 5]*(M2)[ 4]+t_[ 6]*(M2)[ 8]+t_[ 7]*(M2)[ 12]; \
			(M1)[ 5] = t_[ 4]*(M2)[ 1]+t_[ 5]*(M2)[ 5]+t_[ 6]*(M2)[ 9]+t_[ 7]*(M2)[ 13]; \
			(M1)[ 6] = t_[ 4]*(M2)[ 2]+t_[ 5]*(M2)[ 6]+t_[ 6]*(M2)[10]+t_[ 7]*(M2)[ 14]; \
			(M1)[ 7] = t_[ 4]*(M2)[ 3]+t_[ 5]*(M2)[ 7]+t_[ 6]*(M2)[11]+t_[ 7]*(M2)[ 15]; \
			(M1)[ 8] = t_[ 8]*(M2)[ 0]+t_[ 9]*(M2)[ 4]+t_[10]*(M2)[ 8]+t_[11]*(M2)[ 12]; \
			(M1)[ 9] = t_[ 8]*(M2)[ 1]+t_[ 9]*(M2)[ 5]+t_[10]*(M2)[ 9]+t_[11]*(M2)[ 13]; \
			(M1)[10] = t_[ 8]*(M2)[ 2]+t_[ 9]*(M2)[ 6]+t_[10]*(M2)[10]+t_[11]*(M2)[ 14]; \
			(M1)[11] = t_[ 8]*(M2)[ 3]+t_[ 9]*(M2)[ 7]+t_[10]*(M2)[11]+t_[11]*(M2)[ 15]; \
			(M1)[12] = t_[12]*(M2)[ 0]+t_[13]*(M2)[ 4]+t_[14]*(M2)[ 8]+t_[15]*(M2)[ 12]; \
			(M1)[13] = t_[12]*(M2)[ 1]+t_[13]*(M2)[ 5]+t_[14]*(M2)[ 9]+t_[15]*(M2)[ 13]; \
			(M1)[14] = t_[12]*(M2)[ 2]+t_[13]*(M2)[ 6]+t_[14]*(M2)[10]+t_[15]*(M2)[ 14]; \
			(M1)[15] = t_[12]*(M2)[ 3]+t_[13]*(M2)[ 7]+t_[14]*(M2)[11]+t_[15]*(M2)[ 15]; \
	}

// D = S1 * S2
#define MATRIX44_MULTIPLY(D,S1, S2)\
	{ \
		DlMatrix44 t_; \
			t_[ 0] = (S1)[ 0]*(S2)[ 0]+(S1)[ 1]*(S2)[ 4]+(S1)[ 2]*(S2)[ 8]+(S1)[ 3]*(S2)[ 12]; \
			t_[ 1] = (S1)[ 0]*(S2)[ 1]+(S1)[ 1]*(S2)[ 5]+(S1)[ 2]*(S2)[ 9]+(S1)[ 3]*(S2)[ 13]; \
			t_[ 2] = (S1)[ 0]*(S2)[ 2]+(S1)[ 1]*(S2)[ 6]+(S1)[ 2]*(S2)[10]+(S1)[ 3]*(S2)[ 14]; \
			t_[ 3] = (S1)[ 0]*(S2)[ 3]+(S1)[ 1]*(S2)[ 7]+(S1)[ 2]*(S2)[11]+(S1)[ 3]*(S2)[ 15]; \
			t_[ 4] = (S1)[ 4]*(S2)[ 0]+(S1)[ 5]*(S2)[ 4]+(S1)[ 6]*(S2)[ 8]+(S1)[ 7]*(S2)[ 12]; \
			t_[ 5] = (S1)[ 4]*(S2)[ 1]+(S1)[ 5]*(S2)[ 5]+(S1)[ 6]*(S2)[ 9]+(S1)[ 7]*(S2)[ 13]; \
			t_[ 6] = (S1)[ 4]*(S2)[ 2]+(S1)[ 5]*(S2)[ 6]+(S1)[ 6]*(S2)[10]+(S1)[ 7]*(S2)[ 14]; \
			t_[ 7] = (S1)[ 4]*(S2)[ 3]+(S1)[ 5]*(S2)[ 7]+(S1)[ 6]*(S2)[11]+(S1)[ 7]*(S2)[ 15]; \
			t_[ 8] = (S1)[ 8]*(S2)[ 0]+(S1)[ 9]*(S2)[ 4]+(S1)[10]*(S2)[ 8]+(S1)[11]*(S2)[ 12]; \
			t_[ 9] = (S1)[ 8]*(S2)[ 1]+(S1)[ 9]*(S2)[ 5]+(S1)[10]*(S2)[ 9]+(S1)[11]*(S2)[ 13]; \
			t_[10] = (S1)[ 8]*(S2)[ 2]+(S1)[ 9]*(S2)[ 6]+(S1)[10]*(S2)[10]+(S1)[11]*(S2)[ 14]; \
			t_[11] = (S1)[ 8]*(S2)[ 3]+(S1)[ 9]*(S2)[ 7]+(S1)[10]*(S2)[11]+(S1)[11]*(S2)[ 15]; \
			t_[12] = (S1)[12]*(S2)[ 0]+(S1)[13]*(S2)[ 4]+(S1)[14]*(S2)[ 8]+(S1)[15]*(S2)[ 12]; \
			t_[13] = (S1)[12]*(S2)[ 1]+(S1)[13]*(S2)[ 5]+(S1)[14]*(S2)[ 9]+(S1)[15]*(S2)[ 13]; \
			t_[14] = (S1)[12]*(S2)[ 2]+(S1)[13]*(S2)[ 6]+(S1)[14]*(S2)[10]+(S1)[15]*(S2)[ 14]; \
			t_[15] = (S1)[12]*(S2)[ 3]+(S1)[13]*(S2)[ 7]+(S1)[14]*(S2)[11]+(S1)[15]*(S2)[ 15]; \
			MATRIX44_COPY(D, t_); \
	}


#define MATRIX44_BY_VECTOR3(VD, M, VS) \
	{ \
		register float sx_, sy_, sz_; \
			sx_ = (M)[0] * (VS)[0] + (M)[1] * (VS)[1] + (M)[2]  * (VS)[2];\
			sy_ = (M)[4] * (VS)[0] + (M)[5] * (VS)[1] + (M)[6]  * (VS)[2];\
			sz_ = (M)[8] * (VS)[0] + (M)[9] * (VS)[1] + (M)[10] * (VS)[2];\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
			(VD)[2] = sz_;\
	}

#define MATRIX44_BY_NORMAL(VD, M, VS) Matrix44ByNormal((VD), (M), (VS))

#define MATRIX44_BY_NORMAL_W(VD, M, VS) Matrix44ByNormalWithW((VD), (M), (VS))

#define MATRIX44_BY_POINT3(VD, M, VS) \
	{ \
		register float sx_, sy_, sz_; \
			sx_ = (M)[0] * (VS)[0] + (M)[1] * (VS)[1] + (M)[2]  * (VS)[2] + (M)[3];\
			sy_ = (M)[4] * (VS)[0] + (M)[5] * (VS)[1] + (M)[6]  * (VS)[2] + (M)[7];\
			sz_ = (M)[8] * (VS)[0] + (M)[9] * (VS)[1] + (M)[10] * (VS)[2] + (M)[11];\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
			(VD)[2] = sz_;\
	}

#define MATRIX44_BY_POINT3_APPLY_W(VD, M, VS) \
	{ \
		register float sx_, sy_, sz_, swinv_; \
			sx_ = (M)[0] * (VS)[0] + (M)[1] * (VS)[1] + (M)[2]  * (VS)[2] + (M)[3];\
			sy_ = (M)[4] * (VS)[0] + (M)[5] * (VS)[1] + (M)[6]  * (VS)[2] + (M)[7];\
			sz_ = (M)[8] * (VS)[0] + (M)[9] * (VS)[1] + (M)[10] * (VS)[2] + (M)[11];\
			swinv_ = (M)[12] * (VS)[0] + (M)[13] * (VS)[1] + (M)[14]  * (VS)[2] + (M)[15];\
			assert(swinv_ != 0.0f);\
			swinv_ = 1.0f / swinv_;\
			(VD)[0] = sx_*swinv_;\
			(VD)[1] = sy_*swinv_;\
			(VD)[2] = sz_*swinv_;\
	}

#define MATRIX44_BY_POINT3f(VD, M, xx, yy, zz) \
	{ \
		register float sx_, sy_, sz_; \
			sx_ = (M)[0] * xx + (M)[1] * yy + (M)[2]  * zz + (M)[3];\
			sy_ = (M)[4] * xx + (M)[5] * yy + (M)[6]  * zz + (M)[7];\
			sz_ = (M)[8] * xx + (M)[9] * yy + (M)[10] * zz + (M)[11];\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
			(VD)[2] = sz_;\
	}

#define MATRIX44_BY_POINT3f_APPLY_W(VD, M, xx, yy, zz) \
	{ \
		register float sx_, sy_, sz_, swinv_; \
			sx_ = (M)[0] * xx + (M)[1] * yy + (M)[2]  * zz + (M)[3];\
			sy_ = (M)[4] * xx + (M)[5] * yy + (M)[6]  * zz + (M)[7];\
			sz_ = (M)[8] * xx + (M)[9] * yy + (M)[10] * zz + (M)[11];\
			swinv_ = (M)[12] * xx + (M)[13] * yy + (M)[14]  * zz + (M)[15];\
			assert(swinv_ != 0.0f);\
			swinv_ = 1.0f / swinv_;\
			(VD)[0] = sx_*swinv_;\
			(VD)[1] = sy_*swinv_;\
			(VD)[2] = sz_*swinv_;\
	}

#define MATRIX44_BY_POINT3w(VD, o_W, M, VS, i_W) \
	{ \
		register float sx_, sy_, sz_, sw_; \
			sx_ = (M)[0]  * (VS)[0] + (M)[1]  * (VS)[1] + (M)[2]  * (VS)[2] + (M)[3]  * (i_W);\
			sy_ = (M)[4]  * (VS)[0] + (M)[5]  * (VS)[1] + (M)[6]  * (VS)[2] + (M)[7]  * (i_W);\
			sz_ = (M)[8]  * (VS)[0] + (M)[9]  * (VS)[1] + (M)[10] * (VS)[2] + (M)[11] * (i_W);\
			sw_ = (M)[12] * (VS)[0] + (M)[13] * (VS)[1] + (M)[14] * (VS)[2] + (M)[15] * (i_W);\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
			(VD)[2] = sz_;\
			(o_W) = sw_;\
	}

#define MATRIX44_BY_POINT3w_2(VD, o_W, M, VS, i_W) \
	{ \
		register float sx_, sy_; \
			sx_ = (M)[0]  * (VS)[0] + (M)[1]  * (VS)[1] + (M)[2]  * (VS)[2] + (M)[3]  * (i_W);\
			sy_ = (M)[4]  * (VS)[0] + (M)[5]  * (VS)[1] + (M)[6]  * (VS)[2] + (M)[7]  * (i_W);\
			(o_W) = (M)[12] * (VS)[0] + (M)[13] * (VS)[1] + (M)[14] * (VS)[2] + (M)[15] * (i_W);\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
	}
	
#define MATRIX44_BY_POINT3fw(VD, o_W, M, xx, yy, zz, ww) \
	{ \
		register float sx_, sy_, sz_, sw_; \
			sx_ = (M)[0]  * xx + (M)[1]  * yy + (M)[2]  * zz + (M)[3]  * ww;\
			sy_ = (M)[4]  * xx + (M)[5]  * yy + (M)[6]  * zz + (M)[7]  * ww;\
			sz_ = (M)[8]  * xx + (M)[9]  * yy + (M)[10] * zz + (M)[11] * ww;\
			sw_ = (M)[12] * xx + (M)[13] * yy + (M)[14] * zz + (M)[15] * ww;\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
			(VD)[2] = sz_;\
			(o_W) = sw_;\
	}


#define MATRIX44_BY_POINT4(VD, M, VS ) \
	{ \
		register float sx_, sy_, sz_, sw_; \
			sx_ = (M)[0]  * (VS)[0] + (M)[1]  * (VS)[1] + (M)[2]  * (VS)[2] + (M)[3]  * (VS)[3];\
			sy_ = (M)[4]  * (VS)[0] + (M)[5]  * (VS)[1] + (M)[6]  * (VS)[2] + (M)[7]  * (VS)[3];\
			sz_ = (M)[8]  * (VS)[0] + (M)[9]  * (VS)[1] + (M)[10] * (VS)[2] + (M)[11] * (VS)[3];\
			sw_ = (M)[12] * (VS)[0] + (M)[13] * (VS)[1] + (M)[14] * (VS)[2] + (M)[15] * (VS)[3];\
			(VD)[0] = sx_;\
			(VD)[1] = sy_;\
			(VD)[2] = sz_;\
			(VD)[3] = sw_;\
	}

#define MATRIX44_TRANSPOSE(D, S) \
	{ \
		(D)[0] = (S)[0];\
			(D)[1] = (S)[4];\
			(D)[2] = (S)[8];\
			(D)[3] = (S)[12];\
			(D)[4] = (S)[1];\
			(D)[5] = (S)[5];\
			(D)[6] = (S)[9];\
			(D)[7] = (S)[13];\
			(D)[8] = (S)[2];\
			(D)[9] = (S)[6];\
			(D)[10] = (S)[10];\
			(D)[11] = (S)[14];\
			(D)[12] = (S)[3];\
			(D)[13] = (S)[7];\
			(D)[14] = (S)[11];\
			(D)[15] = (S)[15];\
	}

#define MATRIX44_TRANSPOSE_IN_PLACE(S) \
	{ \
		DlMatrix44 t_; \
			MATRIX44_COPY(t_,(S)); \
			MATRIX44_TRANSPOSE((S),t_);\
	}

#define MATRIX44_PRINT(S, Comment) \
	{ \
		printf("\n \t MATRIX44: %s ", (Comment));\
		printf("\n %5f %5f %5f %5f", (S)[0], (S)[1], (S)[2], (S)[3]);\
		printf("\n %5f %5f %5f %5f", (S)[4], (S)[5], (S)[6], (S)[7]);\
		printf("\n %5f %5f %5f %5f", (S)[8], (S)[9], (S)[10], (S)[11]);\
		printf("\n %5f %5f %5f %5f\n\n", (S)[12], (S)[13], (S)[14], (S)[15]);\
	}

#endif				 // __dlMatrix44_H

